{-# LANGUAGE CPP #-}
module Network.Lean.Lai.Handler.Larp (
      LarpRequestState(..)
    , LarpResponseState(..)
    , LarpRequest
    , LarpResponse
    , LarpApplication
    , run
    ) where

import Data.ByteString (ByteString)
import qualified Data.ByteString as B
import qualified Data.ByteString.UTF8 as B
import Control.Exception as E
import Data.Serialize.Get
import qualified Data.Serialize.Ldsf as Ldsf
import qualified Data.Serialize.Ldsf.Types as Ldsf
import Data.Serialize.Ldsf.Lean (composeResponse, getRequest)
import qualified Data.Serialize.Ldsf.Lean.InterfaceFunction as IF
import Data.Serialize.Ldsf.Lean.Category (lookupFunction)
import Data.Streaming.Network (bindPortTCP)
import Network.Lean.Lai
import Network.Lean.Lai.Internal
import Network.Simple.TCP (Socket, SockAddr, HostPreference(..), serve, recv, send)

type Port = Int

data LarpRequestState =
    LarpRequestState
    { larpRequestSocketAddr :: SockAddr
    } deriving (Show) -- only for debugging
data LarpResponseState =
    LarpResponseState
    deriving (Show) -- only for debugging

type LarpRequest = Request LarpRequestState
type LarpResponse = Response LarpResponseState
type LarpApplication = Application LarpRequestState LarpResponseState

-- | Run a 'LarpApplication' on the specified 'Port'.
-- | TODO: test if network-simple handles slow lorris.
run :: Port -> LarpApplication -> IO ()
run p app = serve (Host "127.0.0.1") (show p) $ \(socket, remoteAddr) -> do
    putStrLn $ (show remoteAddr) ++ " established a TCP connection with us."
    processRequest app socket remoteAddr Nothing Nothing

-- | Send a response.
sendResponse :: Socket -> LarpResponse -> IO ResponseReceived
sendResponse socket response = do
    send socket $ composeResponse response
    pure ResponseReceived

-- | Process a single request for the 'Application' using partial
-- | continuation style.
processRequest ::  LarpApplication -> Socket -> SockAddr -> Maybe ByteString ->
          Maybe (ByteString -> Result (Ldsf.Header, LarpRequest)) ->
          IO ()
processRequest app socket socketAddr leftovers continuation = do
    let remote = show socketAddr
    
    -- Maybe we have leftovers, if we don't then we need to receive more data.
    maybeReqdata <- case leftovers of
        Nothing -> recv socket 1024
        Just l  -> pure $ Just l
    
    -- Check if the connection was closed.
    case maybeReqdata of
        Nothing -> do
            putStrLn $ remote ++ " closed the connection."
            pure ()
        Just reqdata  -> do
            -- Time to parse. If we have a continuation then we use it,
            -- otherwise we start a new parse from scratch.
            let parseResult =
                    case continuation of
                        Nothing   -> Ldsf.parseIncremental (getRequest getArguments) reqdata
                        Just cont -> cont reqdata

            case parseResult of
                Fail err bs -> do
                    -- Failed to parse the data from the client. At this point
                    -- we just close the connection and no longer care.
                    putStrLn $ "request from " ++
                        remote ++ " failed with error " ++ err
                    pure ()
                Partial cont ->
                    -- We couldn't parse a complete request so we ended up
                    -- here. We pass the continuation back to ask for more data
                    -- and pass it to the continuation.
                    processRequest app socket socketAddr Nothing (Just cont)
                Done (header, request) bs  -> do
                    -- A complete request was made, awesome. Now we add
                    -- LARP specific state to the request object and pass
                    -- it to users application for handling.
                    let enhancedRequest = request
                            { requestInternal = Just LarpRequestState
                                { larpRequestSocketAddr = socketAddr
                                }
                            }
                    
                    _ <- app enhancedRequest (sendResponse socket)
                    -- Now we ask for more requests if the client wants to
                    -- re-use the socket.
                    processRequest app socket socketAddr (Just bs) Nothing

-- | Parse the arguments of a request.
getArguments :: Ldsf.Version -> LarpRequest -> IF.GetArguments
getArguments ldsfVersion request = 
    case maybeGetArgs of
         Nothing -> error "fixme: function not found"
         Just f  -> IF.getArgs f
    where
        maybeGetArgs = lookupFunction ldsfVersion (requestCategory request)
            (requestVersion request) (requestFunction request)
