module Data.Serialize.Ldsf (
      Magic
    , Version
    , Header(..)
    , latest
    , magicNumber
    , parseIncremental
    , compose
    , composeVersioned
    , getUInt8, getUInt16, getUInt32, getUInt64
    , getInt8, getInt16, getInt32, getInt64
    , getFloat32, getFloat64
    , getByteString8, getByteString16, getByteString32, getByteString64
    , getText8, getText16, getText32, getText64
    , putUInt8, putUInt16, putUInt32, putUInt64
    , putInt8, putInt16, putInt32, putInt64
    , putFloat32, putFloat64
    , putByteString8, putByteString16, putByteString32, putByteString64
    , putText8, putText16, putText32, putText64
    , putValue, putValues
    ) where

import Control.Monad (when)
import Data.ByteString (ByteString)
import qualified Data.ByteString as B
import qualified Data.ByteString.Lazy as BL
import qualified Data.Text as T
import qualified Data.Text.Encoding as T
import qualified Data.Text.Lazy as TL
import qualified Data.Text.Lazy.Encoding as TL
import Data.Serialize.Get (Get, Result, getWord8, getWord16le, getWord32le,
                           getWord64le, getByteString, getLazyByteString,
                           runGetPartial)
import Data.Serialize.IEEE754 (getFloat32le, getFloat64le, putFloat32le,
                               putFloat64le)
import Data.Serialize.Ldsf.Types
import Data.Serialize.Ldsf.Internal
import Data.Serialize.Put (Put, Putter, putWord8, putWord16le, putWord32le,
                           putWord64le, putByteString, putLazyByteString,
                           runPut)

-- | The type of the magic number.
type Magic = UInt16

-- | The type of the LDSF protocol version.
type Version = UInt8

-- | The LDSF header.
data Header = Header
    { headerMagic :: Magic        -- ^ The magic number.
    , headerVersion :: Version    -- ^ The version.
    }
instance Show Header where
    show header = concat
        [ "Magic: ", show $ headerMagic header, ", "
        , "Version: ", show $ headerVersion header
        ]

-- | The current version of the LDSF standard.
latest :: Version
latest = 0

-- | The magic number for our format. It is ACE1 in hex.
magicNumber :: Magic
magicNumber = 57772

-- | Parse LDSF Header with cereals incremental interface.
parseIncremental :: (Version -> Get payload) -> B.ByteString -> Result (Header, payload)
parseIncremental getRequest = runGetPartial (get getRequest)

-- | Read a 'UInt8' from the 'Get' monad state.
getUInt8 :: Get UInt8
getUInt8 = getWord8

-- | Read a 'UInt16' from the 'Get' monad state.
getUInt16 :: Get UInt16
getUInt16 = getWord16le

-- | Read a 'UInt32' from the 'Get' monad state.
getUInt32 :: Get UInt32
getUInt32 = getWord32le

-- | Read a 'UInt64' from the 'Get' monad state.
getUInt64 :: Get UInt64
getUInt64 = getWord64le

-- | Read a 'Int8' from the 'Get' monad state.
getInt8 :: Get Int8
getInt8 = getWord8 >>= \w -> pure $ fromIntegral w

-- | Read a 'Int16' from the 'Get' monad state.
getInt16 :: Get Int16
getInt16 = getWord16le >>= \w -> pure $ fromIntegral w

-- | Read a 'Int32' from the 'Get' monad state.
getInt32 :: Get Int32
getInt32 = getWord32le >>= \w -> pure $ fromIntegral w

-- | Read a 'Int64' from the 'Get' monad state.
getInt64 :: Get Int64
getInt64 = getWord64le >>= \w -> pure $ fromIntegral w

-- | Read a 'Float32' from the 'Get' monad state.
getFloat32 :: Get Float32
getFloat32 = getFloat32le

-- | Read a 'Float64' from the 'Get' monad state.
getFloat64 :: Get Float64
getFloat64 = getFloat64le

-- | Read a 'Text8' from the 'Get' monad state.
getText8 :: Get Text8
getText8 = getByteString8 >>= \bs -> pure $ T.decodeUtf8 bs

-- | Read a 'Text16' from the 'Get' monad state.
getText16 :: Get Text16
getText16 = getByteString16 >>= \bs -> pure $ T.decodeUtf8 bs

-- | Read a 'Text32' from the 'Get' monad state.
getText32 :: Get Text32
getText32 = getByteString32 >>= \bs -> pure $ TL.decodeUtf8 bs

-- | Read a 'Text64' from the 'Get' monad state.
getText64 :: Get Text64
getText64 = getByteString64 >>= \bs -> pure $ TL.decodeUtf8 bs

-- | Read a 'ByteString8' from the 'Get' monad state.
getByteString8 :: Get ByteString8
getByteString8 = getUInt8 >>= \size -> getByteString (fromIntegral size)

-- | Read a 'ByteString16' from the 'Get' monad state.
getByteString16 :: Get ByteString16
getByteString16 = getUInt16 >>= \size -> getByteString (fromIntegral size)

-- | Read a 'ByteString32' from the 'Get' monad state.
getByteString32 :: Get ByteString32
getByteString32 = getUInt32 >>= \size -> getLazyByteString (fromIntegral size)

{-# WARNING getByteString64 "This function only reads up to Int64 length, not UInt64!" #-}
-- | Read a 'ByteString64' from the 'Get' monad state.
-- | WARNING: This function only reads up to Int64 length, not UInt64!
-- | TODO: Implement UInt64 handling.
getByteString64 :: Get ByteString64
getByteString64 = getInt64 >>= \size -> getLazyByteString size

-- | Write a 'UInt8'.
putUInt8 :: Putter UInt8
putUInt8 = putWord8

-- | Write a 'UInt16'.
putUInt16 :: Putter UInt16
putUInt16 = putWord16le

-- | Write a 'UInt32'.
putUInt32 :: Putter UInt32
putUInt32 = putWord32le

-- | Write a 'UInt64'.
putUInt64 :: Putter UInt64
putUInt64 = putWord64le

-- | Write a 'Int8'.
putInt8 :: Putter Int8
putInt8 = putWord8 . fromIntegral

-- | Write a 'Int16'.
putInt16 :: Putter Int16
putInt16 = putWord16le . fromIntegral

-- | Write a 'Int32'.
putInt32 :: Putter Int32
putInt32 = putWord32le . fromIntegral

-- | Write a 'Int64'.
putInt64 :: Putter Int64
putInt64 = putWord64le . fromIntegral

-- | Write a IEEE754 'Float32'.
putFloat32 :: Putter Float32
putFloat32 = putFloat32le

-- | Write a IEEE754 'Float64'.
putFloat64 :: Putter Float64
putFloat64 = putFloat64le

-- | Write the 'UInt8' size-prefix and the associated 'ByteString'.
putByteString8 :: Putter ByteString8
putByteString8 bs = do
    putUInt8 size
    putByteString $ B.take (fromIntegral size) bs
    where
        size = fitBounded (B.length bs) :: UInt8

-- | Write the 'UInt16' size-prefix and the associated 'ByteString'.
putByteString16 :: Putter ByteString16
putByteString16 bs = do
    putUInt16 size
    putByteString $ B.take (fromIntegral size) bs
    where
        size = fitBounded (B.length bs) :: UInt16

-- | Write the 'UInt32' size-prefix and the associated 'ByteString'.
putByteString32 :: Putter ByteString32
putByteString32 bs = do
    putUInt32 size
    putLazyByteString $ BL.take (fromIntegral size) bs
    where
        size = fitBounded (BL.length bs) :: UInt32

{-# WARNING putByteString64 "This function only writes up to Int64 length, not UInt64!" #-}
-- | Write the 'UInt64' size-prefix and the associated 'ByteString'.
-- | WARNING: This function only writes up to Int64 length, not UInt64!
-- | TODO: Implement UInt64 handling.
putByteString64 :: Putter ByteString64
putByteString64 bs = do
    putInt64 size
    putLazyByteString $ BL.take size bs
    where
        size = fitBounded (BL.length bs) :: Int64

-- | Write the 'UInt8' size-prefix and the associated 'Text'.
putText8 :: Putter Text8
putText8 = putByteString8 . T.encodeUtf8

-- | Write the 'UInt16' size-prefix and the associated 'Text'.
putText16 :: Putter Text16
putText16 = putByteString16 . T.encodeUtf8

-- | Write the 'UInt32' size-prefix and the associated 'Text'.
putText32 :: Putter Text32
putText32 = putByteString32 . TL.encodeUtf8

-- | Write the 'UInt64' size-prefix and the associated 'Text'.
putText64 :: Putter Text64
putText64 = putByteString64 . TL.encodeUtf8

-- | Write a 'Value'.
putValue :: Putter Value
putValue value =
    case value of
         ValueNothing        -> pure ()
         ValueUInt8 v        -> putUInt8 v
         ValueUInt16 v       -> putUInt16 v
         ValueUInt32 v       -> putUInt32 v
         ValueUInt64 v       -> putUInt64 v
         ValueInt8 v         -> putInt8 v
         ValueInt16 v        -> putInt16 v
         ValueInt32 v        -> putInt32 v
         ValueInt64 v        -> putInt64 v
         ValueFloat32 v      -> putFloat32 v
         ValueFloat64 v      -> putFloat64 v
         ValueText8 v        -> putText8 v
         ValueText16 v       -> putText16 v
         ValueText32 v       -> putText32 v
         ValueText64 v       -> putText64 v
         ValueByteString8 v  -> putByteString8 v
         ValueByteString16 v -> putByteString16 v
         ValueByteString32 v -> putByteString32 v
         ValueByteString64 v -> putByteString64 v

-- | Write a list '[]' of 'Value'.
putValues :: Putter [Value]
putValues = mapM_ putValue

-- | Read the LDSF header with user supplied payload getter.
get :: (Version -> Get request) -> Get (Header, request)
get getRequest = do
    magic <- getUInt16
    when (magic /= magicNumber)
         (fail $ "Expected magic number " ++ show magicNumber ++ ", received " ++ show magic)
    version <- getUInt8
    request <- getRequest version
    pure (Header
        { headerMagic = magic
        , headerVersion = version
        }, request)

-- | Compose the latest version of the LDSF standard as a 'ByteString' with
-- | user supplied payload put function.
-- | 
-- | WARNING: For API stability you may want to use 'composeVersioned'
-- |          instead and supply a specific LDSF version number.
compose :: (Version -> Put) -> ByteString
compose = composeVersioned latest

-- | Compose an LDSF structure as a 'ByteString' with user supplied version and
-- | payload put function.
composeVersioned :: Version -> (Version -> Put) -> ByteString
composeVersioned version putPayload = runPut (putVersioned version putPayload)

-- | Write the latest version of the LDSF standard with user supplied payload
-- | put function.
-- | 
-- | WARNING: For API stability you may want to use 'putVersioned'
-- |          instead and supply a specific LDSF version number.
put :: (Version -> Put) -> Put
put = putVersioned latest

-- | Write the versioned LDSF structure with user supplied payload put function.
putVersioned :: Version -> (Version -> Put) -> Put
putVersioned version putPayload = do
    putUInt16 magicNumber
    putUInt8 version
    putPayload version
