{-# LANGUAGE TemplateHaskell   #-}
module Main where

import Control.Monad
import Data.Either (isRight)
import Data.Serialize.Get (Get, runGet)
import Data.Serialize.Put (Putter, runPut)
import Data.Serialize.Ldsf
import Data.Serialize.Ldsf.Types
import Data.Serialize.Ldsf.Internal
import System.Exit
import Test.QuickCheck
import Test.QuickCheck.Instances ()

-- | Given a putter and a getter, check if they are reversible.
reversible :: a -> Putter a -> Get a -> Bool
reversible v put get = isRight $ runGet get $ runPut (put v)

-- test internals first to make sure they work

-- Tests if an 'Integer' is between a bounded range.
fitBoundedBetween :: Integer -> Integer -> Integer -> Integer -> Bool
fitBoundedBetween x x' minValue maxValue
    | x < minValue = x' == minValue
    | x > maxValue = x' == maxValue
    | otherwise    = x' >= minValue && x' <= maxValue

-- NB: we don't need to property test all types fitBounded, just:
-- large to small (signed and unsigned)
-- small to large (signed and unsigned)

-- Int32 -> UInt8
prop_fitBounded_int32_to_uint8 :: Int32 -> Bool
prop_fitBounded_int32_to_uint8 x = fitBoundedBetween original bounded  minValue maxValue
    where
        original = fromIntegral x :: Integer
        bounded = fromIntegral ((fitBounded x) :: UInt8) :: Integer
        minValue = fromIntegral (minBound :: UInt8) :: Integer
        maxValue = fromIntegral (maxBound :: UInt8) :: Integer

-- Int32 -> Int8
prop_fitBounded_int32_to_int8 :: Int32 -> Bool
prop_fitBounded_int32_to_int8 x = fitBoundedBetween original bounded minValue maxValue
    where
        original = fromIntegral x :: Integer
        bounded = fromIntegral ((fitBounded x) :: Int8) :: Integer
        minValue = fromIntegral (minBound :: Int8) :: Integer
        maxValue = fromIntegral (maxBound :: Int8) :: Integer

-- check if it works from small to large too

-- Int32 -> Int64
prop_fitBounded_int32_to_int64 :: Int32 -> Bool
prop_fitBounded_int32_to_int64 x = fitBoundedBetween original bounded  minValue maxValue
    where
        original = fromIntegral x :: Integer
        bounded = fromIntegral ((fitBounded x) :: Int64) :: Integer
        minValue = fromIntegral (minBound :: Int64) :: Integer
        maxValue = fromIntegral (maxBound :: Int64) :: Integer

-- Int32 -> UInt64
prop_fitBounded_int32_to_uint64 :: Int32 -> Bool
prop_fitBounded_int32_to_uint64 x = fitBoundedBetween original bounded  minValue maxValue
    where
        original = fromIntegral x :: Integer
        bounded = fromIntegral ((fitBounded x) :: UInt64) :: Integer
        minValue = fromIntegral (minBound :: UInt64) :: Integer
        maxValue = fromIntegral (maxBound :: UInt64) :: Integer

prop_splitcombine_int8 :: Integer -> Bool
prop_splitcombine_int8 x = x == (combineInteger $ (splitInteger x :: [Int8]))

prop_splitcombine_int16 :: Integer -> Bool
prop_splitcombine_int16 x = x == (combineInteger $ (splitInteger x :: [Int16]))

prop_splitcombine_int32 :: Integer -> Bool
prop_splitcombine_int32 x = x == (combineInteger $ (splitInteger x :: [Int32]))

prop_splitcombine_int64 :: Integer -> Bool
prop_splitcombine_int64 x = x == (combineInteger $ (splitInteger x :: [Int64]))

-- Now test publicly accessible functions.

prop_uint8_reversible :: UInt8 -> Bool
prop_uint8_reversible v = reversible v (putUInt8) (getUInt8)

prop_uint16_reversible :: UInt16 -> Bool
prop_uint16_reversible v = reversible v (putUInt16) (getUInt16)

prop_uint32_reversible :: UInt32 -> Bool
prop_uint32_reversible v = reversible v (putUInt32) (getUInt32)

prop_uint64_reversible :: UInt64 -> Bool
prop_uint64_reversible v = reversible v (putUInt64) (getUInt64)

prop_int8_reversible :: Int8 -> Bool
prop_int8_reversible v = reversible v (putInt8) (getInt8)

prop_int16_reversible :: Int16 -> Bool
prop_int16_reversible v = reversible v (putInt16) (getInt16)

prop_int32_reversible :: Int32 -> Bool
prop_int32_reversible v = reversible v (putInt32) (getInt32)

prop_int64_reversible :: Int64 -> Bool
prop_int64_reversible v = reversible v (putInt64) (getInt64)

prop_float32_reversible :: Float32 -> Bool
prop_float32_reversible v = reversible v (putFloat32) (getFloat32)

prop_float64_reversible :: Float64 -> Bool
prop_float64_reversible v = reversible v (putFloat64) (getFloat64)

prop_text8_reversible :: Text8 -> Bool
prop_text8_reversible v = reversible v (putText8) (getText8)

prop_text16_reversible :: Text16 -> Bool
prop_text16_reversible v = reversible v (putText16) (getText16)

prop_text32_reversible :: Text32 -> Bool
prop_text32_reversible v = reversible v (putText32) (getText32)

prop_text64_reversible :: Text64 -> Bool
prop_text64_reversible v = reversible v (putText64) (getText64)

prop_bytestring8_reversible :: ByteString8 -> Bool
prop_bytestring8_reversible v = reversible v (putByteString8) (getByteString8)

prop_bytestring16_reversible :: ByteString16 -> Bool
prop_bytestring16_reversible v = reversible v (putByteString16) (getByteString16)

prop_bytestring32_reversible :: ByteString32 -> Bool
prop_bytestring32_reversible v = reversible v (putByteString32) (getByteString32)

prop_bytestring64_reversible :: ByteString64 -> Bool
prop_bytestring64_reversible v = reversible v (putByteString64) (getByteString64)

-- | Use Template Haskell to automatically run all of the above properties.
return []
runTests :: IO Bool
runTests = $quickCheckAll

main :: IO ()
main = do
    result <- runTests
    unless result exitFailure
